/* Panorama_Tools	-	Generate, Edit and Convert Panoramic Images
   Copyright (C) 1998,1999,2000 - Helmut Dersch  der@fh-furtwangen.de
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*------------------------------------------------------------*/

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>
#include "filter.h"

void 	PV_ExtractStill( TrformStr *TrPtr );
void 	PV_SetInvMakeParams( struct fDesc *stack, struct MakeParams *mp, Image *im , Image *pn );
int 	SetUpAtan();
int 	SetUpSqrt();
int 	SetUpMweights();

void 	DrawWindow();
void 	DrawView(int InterPolator);

int readJPEG ( Image *im, fullPath *sfile );

int					panning = FALSE;
int					zooming_in = FALSE;
int					zooming_out = FALSE;

#define VIEW_WIDTH	400
#define VIEW_HEIGHT 300


char text[] = "PTViewer";

Image 			view, pano;
Display 		*disp;
Window 			win;
int 			screen;
XImage 			*ximage = NULL;
GC 				gc;
Visual    		*visual;
unsigned int  	depth;
int 			oldposx, oldposy;


int main( int argc, char** argv )
{
	XSizeHints hint;
	unsigned long fg, bg;
	int done;
 	union 
 	{
    	XEvent              event;
    	XAnyEvent           any;
    	XButtonEvent        button;
    	XKeyEvent           key;
    	XConfigureEvent     configure;
    	XExposeEvent        expose;
    	XMotionEvent        motion;
    	XResizeRequestEvent resize;
    	XClientMessageEvent message;
 	 } event;
  	int newposx, newposy;
  	static Atom           	proto_atom= None, delete_atom= None;


	// Set up display
	


	disp 		= XOpenDisplay("");
	screen 		= DefaultScreen(disp);
  	depth		= DefaultDepth(disp, screen);
   	
   	if( depth != 24 )
   	{
   		PrintError("Depth = %d, must be 24 pixels", (int)depth);
   	}
 	bg = WhitePixel(disp,screen);
	fg = BlackPixel(disp,screen);

 	// Load panoramic image  	
	if( argc < 2 )
	{
		PrintError("No image file supplied");
		exit(0);
	}
	
	if( strcmp( argv[1], "-h" ) == 0 )
	{
		PrintError( "%s\n %s\nUsage: PTViewer Imagefile", "PTViewer", LONGVERSION );
		exit(0);
	}
	if( readJPEG( &pano, (fullPath*)argv[1] ) != 0 )
	{
		PrintError("Could not read panoramic image");
		exit(0);
	}
	pano.hfov 		= 360.0;
	pano.format 	= _equirectangular;
	
	// Initialize Look-up tables
	
	if( SetUpAtan() || SetUpSqrt() || SetUpMweights() )
	{
		PrintError("Could not set up LU Tables");
		exit(0);
	}

	// Set up viewer window

	SetImageDefaults( &view );

	view.hfov 			= 70.0;
	view.width 			= VIEW_WIDTH;
	view.height			= VIEW_HEIGHT;
	view.bitsPerPixel 	= 32;
	view.bytesPerLine 	= view.width * view.bitsPerPixel / 8;
	view.dataSize 		= view.bytesPerLine * view.height;
	view.format			= 1;
	view.data 			= (unsigned char**)mymalloc( view.dataSize );
	if(view.data == NULL)
	{
		PrintError("Not enough memory");
		exit(0);
	}


	hint.x 			= 200;
	hint.y 			= 200;
	hint.width 		= view.width;
	hint.height 	= view.height;
	hint.flags=PPosition | PSize;

	win = XCreateSimpleWindow(disp,
		DefaultRootWindow(disp),
		hint.x,
		hint.y,
		hint.width,
		hint.height,
		5,
		fg,
		bg);

	XSetStandardProperties(disp,
		win,
		text,
		text,
		None,
		argv,
		argc,
		&hint);
	gc = XCreateGC(disp,win,0,0);
	XSetBackground(disp,gc,bg);
	XSetForeground(disp,gc,fg);

	XSelectInput(disp,
		win,
		ButtonPressMask | ButtonReleaseMask | Button1MotionMask | KeyPressMask | KeyReleaseMask |
    	StructureNotifyMask | EnterWindowMask | LeaveWindowMask| ExposureMask);
	XMapRaised(disp,win);

    proto_atom 	= XInternAtom(disp, "WM_PROTOCOLS", False);
    delete_atom = XInternAtom(disp, "WM_DELETE_WINDOW", False);
    if ((proto_atom != None) && (delete_atom != None))
   		XChangeProperty(disp, win, proto_atom, XA_ATOM, 32, PropModeReplace, (unsigned char *)&delete_atom, 1);
 


   	visual  = DefaultVisual(disp, screen);

   	ximage = XCreateImage(disp, visual, depth, ZPixmap, 0,
                                          NULL, view.width, view.height,
                                          8, 0);
    memset( *(view.data), 0, view.dataSize);
    ximage->data = (char *)*(view.data);
    ximage->byte_order= MSBFirst;

	

	done = 0;
	while( done==0)
	{
		int isChanged = 0;
		
		// No XNextEvent 
				
		if( XCheckTypedEvent(disp, Expose, (XEvent*)&event) ) DrawWindow();

		if( XCheckTypedEvent(disp, KeyPress, (XEvent*)&event) )
		{
      		char buf[128];
      		KeySym ks;
      		XComposeStatus status;
      		int nkey = XLookupString(&event.key,buf,128,&ks,&status);
      						
      		if( nkey == 0 )
      		{
        		switch(ks)
      			{
      				case XK_Shift_L:
					case XK_Shift_R: zooming_in = TRUE;
 					break;
      				
      				case XK_Control_L:
					case XK_Control_R: zooming_out = TRUE;
      				break;
      			}
      		}
      		else if( nkey == 1 && buf[0] == 'q')
     			done = 1;
     						
     		goto _EventLoop;
      	}

		if( XCheckTypedEvent(disp, KeyRelease, (XEvent*)&event) ) 
		{
      		char buf[128];
      		KeySym ks;
      		XComposeStatus status;

      		if (XLookupString(&event.key,buf,128,&ks,&status) == 0)
      		{
      			switch(ks)
      			{
      				case XK_Shift_L:
                	case XK_Shift_R: zooming_in = FALSE;
      								view.format = 1;
      								DrawWindow();
      								break;
      				case XK_Control_L: 
	                case XK_Control_R: zooming_out = FALSE;
      								view.format = 1;
      								DrawWindow();
      								break;
      			}
      		}
     		goto _EventLoop;
		}
 
 		if( XCheckTypedEvent(disp, MotionNotify, (XEvent*)&event) )
 		{ 
 			newposx = event.button.x;
            newposy = event.button.y;
    		while (XCheckTypedEvent(disp, MotionNotify, (XEvent*)&event) == True) 
    		{
				newposx= event.button.x;
				newposy= event.button.y;
     		}
		}
					
		if( XCheckTypedEvent(disp, ButtonPress, (XEvent*)&event) )
		{
      		if (event.button.button == 1) 
      		{
      			panning = TRUE;
				oldposx = newposx = event.button.x;
				oldposy = newposy = event.button.y;
			}
		}

		if( XCheckTypedEvent(disp, ButtonRelease, (XEvent*)&event) )
		{
      		if (event.button.button == 1) 
      		{
				panning = FALSE;
				view.format = 1;
				DrawWindow();
			}
		}
		
		if( XCheckTypedEvent(disp, DestroyNotify, (XEvent*)&event) )
			done = 1;
   		
		if( XCheckTypedEvent(disp,ClientMessage , (XEvent*)&event) )
		{
			if ((event.message.window == win) && (event.message.data.l[0] == delete_atom)) 
				done = 1;
		}

		
_EventLoop:			
		if( panning )
		{
				double yaw, pitch;
										
				yaw = view.yaw + (newposx - oldposx)/20.0;
				NORM_ANGLE( yaw );
										
				pitch = view.pitch - (newposy - oldposy)/20.0;
				if( pitch > 90.0 ) pitch = 90.0;
				if( pitch < -90.0 ) pitch = -90.0;
									
				if( pitch != view.pitch || yaw != view.yaw )
				{
					view.pitch 	= pitch;
					view.yaw 	= yaw;
					isChanged = 1;
				}
		}
		if( zooming_in && view.hfov > 10.5)
		{
			view.hfov /= 1.03;
			isChanged = 1;
		}
		if( zooming_out && view.hfov < 165.0)
		{
			view.hfov *= 1.03;
			isChanged = 1;
		}
		if( zooming_in || zooming_out || panning )
		{
			if( isChanged )
			{
				view.format = 1;
				DrawWindow();
			}
		}


	}
	XFreeGC(disp,gc);
	XDestroyWindow(disp, win);
	XCloseDisplay(disp);

	return (0);
}
	





// Error reporting

void  PrintError( char* fmt, ...)
{
	va_list	ap;
	char message[257];
	
	va_start(ap, fmt);
	vsprintf(message, fmt, ap);
	va_end(ap);
	
	printf("%s\n", message);
}

void**  mymalloc( long numBytes )					// Memory allocation, use Handles
{
	char **mem;
	
	mem = (char**)malloc( sizeof(char*) );			// Allocate memory for pointer
	if(mem == NULL)
		return (void**)NULL;
	else
	{
		(*mem) = (char*) malloc( numBytes );		// Allocate numBytes
		if( *mem == NULL )
		{
			free( mem );
			return (void**)NULL;
		}
		else
			return (void**)mem;
	}
}

void 	myfree( void** Hdl )						// free Memory, use Handles
{
	free( (char*) *Hdl );
	free( (char**) Hdl );
}		


void SetImageDefaults(Image *im)
{
	im->data 			= NULL;
	im->bytesPerLine	= 0;
	im->width			= 0;
	im->height			= 0;
	im->dataSize		= 0;
	im->bitsPerPixel	= 0;
	im->format			= 0;
	im->dataformat		= _RGB;
	im->hfov			= 0.0;
	im->yaw				= 0.0;
	im->pitch			= 0.0;
	im->roll			= 0.0;
//	SetCorrectDefaults( &(im->cP) );
	*(im->name)			= 0;
}

// expand image from 3 to 4 bits per pixel. No pad bytes allowed.
// Memory must be allocated
void ThreeToFourBPP( Image *im )
{
	register int x,y,c1,c2;

	if( im->bitsPerPixel == 32 || im->bitsPerPixel == 64) // Nothing to do
		return;
	
	
	
	if( im->bitsPerPixel == 24 )	// Convert to 4byte / pixel
	{
		for( y = im->height-1; y>=0; y--)
		{
			for( x= im->width-1; x>=0; x--)
			{
				c1 = (y * im->width + x) * 4;
				c2 = y * im->bytesPerLine + x * 3;
				(*(im->data))[c1++] = UCHAR_MAX;
				(*(im->data))[c1++] = (*(im->data))[c2++];
				(*(im->data))[c1++] = (*(im->data))[c2++];
				(*(im->data))[c1++] = (*(im->data))[c2++];
			}
		}
		im->bitsPerPixel = 32;
		im->bytesPerLine = im->width * 4;
	}
	else if( im->bitsPerPixel == 48 ) // Convert to 8byte / pixel
	{
		for( y = im->height-1; y>=0; y--)
		{
			for( x= im->width-1; x>=0; x--)
			{
				c1 = (y * im->width + x) * 4;
				c2 = y * im->bytesPerLine/2 + x * 3;
				((USHORT*)(*(im->data)))[c1++] = USHRT_MAX;
				((USHORT*)(*(im->data)))[c1++] = ((USHORT*)(*(im->data)))[c2++];
				((USHORT*)(*(im->data)))[c1++] = ((USHORT*)(*(im->data)))[c2++];
				((USHORT*)(*(im->data)))[c1++] = ((USHORT*)(*(im->data)))[c2++];
			}
		}
		im->bitsPerPixel = 64;
		im->bytesPerLine = im->width * 8;
	}
	im->dataSize = im->height * im->bytesPerLine;
}

static int copy_rgb_to_zpixmap (XImage *dest, const Image *src)
{
	uint32_t *src_data;

	uint32_t src_r;
	uint32_t src_g;
	uint32_t src_b;

	uint32_t dest_r;
	uint32_t dest_g;
	uint32_t dest_b;
	uint32_t dest_pixel_value;

	uint32_t x;
	uint32_t y;
	uint32_t pixel;

	uint32_t dest_r_maxval;
	uint32_t dest_g_maxval;
	uint32_t dest_b_maxval;

	uint32_t dest_r_offset;
	uint32_t dest_g_offset;
	uint32_t dest_b_offset;

	dest_r_offset = 0;
	dest_g_offset = 0;
	dest_b_offset = 0;
	dest_r_maxval = dest->red_mask;
	dest_g_maxval = dest->green_mask;
	dest_b_maxval = dest->blue_mask;
	for (x = 0; x < dest->depth; x++)
	{
		if ((dest_r_maxval & 0x01) == 0)
		{
			dest_r_offset++;
			dest_r_maxval >>= 1;
		}
		if ((dest_g_maxval & 0x01) == 0)
		{
			dest_g_offset++;
			dest_g_maxval >>= 1;
		}
		if ((dest_b_maxval & 0x01) == 0)
		{
			dest_b_offset++;
			dest_b_maxval >>= 1;
		}
	}

	src_data = (uint32_t *) *src->data;

	pixel = 0;
	for (y = 0; y < dest->height; y++)
	{
		for (x = 0; x < dest->width; x++, pixel++)
		{
			int32_t bytenum;

			src_r = (src_data[pixel] >>  8) & 0xFF;
			src_g = (src_data[pixel] >> 16) & 0xFF;
			src_b = (src_data[pixel] >> 24) & 0xFF;

			dest_r = dest_r_maxval * src_r / 0xFF;
			dest_g = dest_g_maxval * src_g / 0xFF;
			dest_b = dest_b_maxval * src_b / 0xFF;

			dest_pixel_value = 0
				| ((dest_r << dest_r_offset) & dest->red_mask)
				| ((dest_g << dest_g_offset) & dest->green_mask)
				| ((dest_b << dest_b_offset) & dest->blue_mask);

			if ((y == 150) && (x == 200))
				printf ("[200x150] RGB(%u,%u,%u) -> RGB(%u,%u,%u) -> 0x%04x\n",
						src_r, src_g, src_b,
						dest_r, dest_g, dest_b,
						dest_pixel_value);
			for (bytenum = 0; bytenum < (dest->depth / 8); bytenum++)
			{
				int offset = (pixel * dest->bits_per_pixel / 8) + bytenum;
				dest->data[(pixel * dest->bits_per_pixel / 8) + bytenum] = 
					(dest_pixel_value >> (dest->bits_per_pixel - (8 * (bytenum + 1)))) & 0xFF;
				if ((y == 150) && (x == 200))
					printf ("Setting byte #%i to 0x%02x\n",
							offset, dest->data[offset]);
			}
		}
	}

	return (0);
} /* int copy_rgb_to_zpixmap */

void DrawWindow()
{
	XWindowAttributes xa;

	XGetWindowAttributes( disp, win, &xa );

	if( xa.width != view.width || xa.height != view.height )
	{
		myfree((void**)view.data);
		view.width 			= xa.width;
		view.height 		= xa.height;
        view.bytesPerLine 	= view.width * view.bitsPerPixel / 8;
        view.dataSize     	= view.bytesPerLine * view.height;
        view.format    		= 1;
        view.data   		= (unsigned char**)mymalloc( view.dataSize );
        if(view.data == NULL)
        {
             PrintError("Not enough memory");
             exit(0);
        }

        ximage = XCreateImage(disp, visual, depth, ZPixmap, 0,
                                          NULL, view.width, view.height,
                                          8, 0);
     	ximage->data = (char *)*(view.data);
    	ximage->byte_order= MSBFirst;
	}
	if( view.format )
	{
		if( panning || zooming_in || zooming_out)
			DrawView(_nn);
		else
			DrawView(_bilinear);
	}

	ximage->data = (char *) malloc (view.dataSize);
	if (ximage->data == NULL)
		return;
	memcpy (ximage->data, *(view.data), view.dataSize);

	copy_rgb_to_zpixmap (ximage, &view);

	XPutImage(disp, win, gc, ximage, 0, 0, 0, 0, view.width, view.height);

	free (ximage->data);
	ximage->data = NULL;
}

void DrawView(int InterPolator)
{
	TrformStr  				Tr;
			
	Tr.interpolator			= InterPolator;

	Tr.src					= &pano;
	Tr.dest					= &view;	
	
	PV_ExtractStill( &Tr );
	
	view.format = 0;
	return;
	
}
		
